
importSchema =
function(node, baseURL)
{
   u =  xmlGetAttr(node, "schemaLocation", NA)
   if(is.na(u))
     return(NULL)

   u = getRelativeURL(u, baseURL)
   parseSchemaDoc(u)
}

parseSchemaDoc =
  # "http://www.ncbi.nlm.nih.gov/entrez/eutils/soap/v2.0/efetch_pubmed.wsdl"
  #
  # Use to use the following two lines:
  #  (with handlers = WSDLParseHandlers(fileName))
  #  library(XML)
     # Parse the WSDL file into a tree which we will pick apart.
     # Could put handlers to convert the elements on the fly.
     #XXX Use internal nodes with xmlInternalTreeParse
  # doc = xmlRoot(xmlTreeParse(fileName, handlers = handlers,  asTree = TRUE, fullNamespaceInfo = TRUE))
  #
function(url, removeComments = TRUE, namespaces = c(xs = "http://www.w3.org/2001/XMLSchema"),
           followImports = TRUE, followIncludes = followImports)
{
  baseURL = dirname(url)
  doc = xmlInternalTreeParse(url)

      # Force the prefix of the namespaces to be xs
  if(length(namespaces) == 0 || !("xs" %in% names(namespaces)))
      names(namespaces)[1] = "xs"

  if(followImports) {
     imports = getNodeSet(doc, "//xs:schema/xs:import", namespaces)

     imports = lapply(imports, function(node) {
                              xdoc = importSchema(node, baseURL)
                              if(is.null(xdoc)) {
                                 removeNodes(node)
                                 return(NULL)
                              }
                              schema = getNodeSet(xdoc, "//xs:schema", namespaces)
                              sapply(schema, function(s) replaceNodes(node, xmlRoot(s)))
                            })
  }

  if(followIncludes) {

    includes = getNodeSet(doc, "//xs:schema/xs:include", namespaces)
    if(length(includes)) {
        sapply(includes, function(node) {
                             xdoc = importSchema(node, baseURL)
                             schema = getNodeSet(xdoc, "//xs:schema", c(xs = "http://www.w3.org/2001/XMLSchema"))
                             p = xmlParent(node)
                             sapply(schema, function(s) addChildren(p, kids = xmlChildren(s)))
                         })
        removeNodes(includes, TRUE)
      }
  }

  if(removeComments) {
     comments = getNodeSet(doc, "//comment()", c(xs="http://www.w3.org/2001/XMLSchema"), noMatchOkay = TRUE)
     if(length(comments))
        removeNodes(comments)
   }
  
  doc
}
